;Software for a lock-opener initiated by knocking the right way.
;This software is distributed under the GNU GPL. See 
;http://www.gnu.org/licenses/gpl.txt for more info. 
;Copyright 2007 J. Domburg <projectknock2open@spritesmods.com>


;We're using a tiny2313 here.
.device ATTiny2313
;You might want to include the ATTiny2313-definitions included with
;your assembler. I use gavrasm, which doesn't need them.
;.include "tn2313def.h"

;The fuses should be configured as follows:
;FUSE_L=0xE4
;FUSE_H=0xdf

;Port declarations
;PortB: Stepper motor drivers
.equ m1		=	2
.equ m2		=	4
.equ m3		=	3
.equ m4		=	5

;portD
.equ left	=	5 ;button
.equ knock	=	4 ;amp output
.equ right	=	2 ;button
.equ led1	=	1
.equ led2	=	0

.def dat	=	r0
.def temp	=	R16 	;The temp register
.def temp2	=	R17 	;Second temp register
.def temp3	=	R18
.def temp4	=	R19
.def isopen	=	r20
.def closectr	=	r21
.def tone	=	r22
.def ttwo	=	r23
.def zlo	=	r30
.def zhi	=	r31


	rjmp start		; reset vector
	reti			; External Int 0 not enabled
	reti			; External Int 1 not enabled
	reti			; Timer 1 Capture Int not enabled
	reti			; Timer 1 Compare match A Int not enabled
	rjmp t1int			; Timer 1 Overflow Int not enabled
	reti			; Timer 0 Overflow Int not enabled
	reti			; UART ReceiveInt
	reti			; UART Date Reg Empty Int not enabled
	reti			; UART tx'ed Int Not enabled
	reti			; Anacomp interrupt Not enabled


start:
	ldi temp,RAMEND		; top of memory
	out SPL,temp		; init stack pointer
	ldi temp,0b111100		; Port B direction options
	out DDRB,temp		; Setup port direction
	ldi temp,0b000011	; Port D direction options
	out DDRD,temp		; Setup port direction
	ldi temp,0b110100	; Port D direction options
	out PORTD,temp		; Setup port direction
	ldi isopen,0

	;Timer1 ints once every 8.4 seconds
	ldi temp,0
	out tccr1a,temp
	ldi temp,0b101
	out tccr1b,temp
	ldi temp,$80
	out timsk,temp

	;We jump here if a false hit is detected. Temp indicates if we can
	;just go ahead and wait for the next knock (0) or if we should wait a 
	;while (!0)
falsehit:
	cpi temp,0
	breq loop
	;Wait a small amount of time if a false hit is detected
	sbi portd,led2
	sbi portd,led1
	ldi temp2,5
falsewait:
	rcall getknock
	cpi temp,0
	brne falsehit
	rcall wait2

	dec temp2
	brne falsewait
	
	sei

;Main loop.	
loop:
	sbi portd,led1
	cbi portd,led2
	ldi temp,0
	out portb,temp
	sbic pind,knock
	 rjmp doknock
	sbis pind,left
	 rcall motorleft
	sbis pind,right
	 rcall motorright
	rjmp loop


;Someone knocked.
doknock:
	cbi portd,led1
	rcall getknock ;Wait for the knock to end
	cpi temp,1     ;Check for erronous knock
	brne falsehit 
	

	sbi portd,led2

	;Measure the duration of the pause between the first and second knock.
	ldi tone,4
knockmeaspause:
	rcall wait2
	rcall getknock
	cpi temp,1
	breq gotpause
	cpi temp,2
	breq falsehit
	
	inc tone
	ldi temp,2
	cpi tone,255
	breq falsehit
	rjmp knockmeaspause
	
	;We;ve got the duration.
gotpause:
	cbi portd,led2

	ldi temp,2
	
	cpi tone,20 ;Check bounds on pause duration
	brlo falsehit


	;Measure pause between 2nd and 3rd knock
	ldi ttwo,4
knockmeaspause2:
	rcall wait2
	rcall getknock
	cpi temp,1
	breq gotpause2
	cpi temp,2
	breq loop
	inc ttwo
	cpi ttwo,255
	breq falsehit
	rjmp knockmeaspause2
gotpause2:

	sbi portd,led2

	;Tone contains first pause, ttwo contains second
	;Check knock - short pause - knock - long pause - knock
	mov temp4,tone
	mov temp2,ttwo
	lsr temp2
	sub temp4,temp2
	lsr temp2
	mov temp,temp2
	lsr temp
	add temp4,temp
	cp temp4,temp2
	brsh notshortlong
	ldi temp,80
	rjmp openloop
	
notshortlong:
	;Check knock - long pause - knock - short pause - knock
	mov temp4,tone
	mov temp2,ttwo
	lsr temp4
	sub temp4,temp2
	lsr temp2
	mov temp,temp2
	lsr temp
	add temp4,temp
	cp temp4,temp2
	brsh notlongshort
	ldi temp,80
	rjmp closeloop
	
notlongshort:
	;None of the above. Error out.
	ldi temp,2
	rjmp falsehit
	

;Open the lock
openloop:
	push temp
	rcall motorleft
	pop temp
	dec temp
	brne openloop
	ldi temp,0
	out portb,temp
	rcall wait
	rcall wait
	ldi isopen,1
	rjmp loop

;Close the lock
closeloop:
	push temp
	rcall motorright
	pop temp
	dec temp
	brne closeloop
	ldi temp,0
	out portb,temp
	rcall wait
	rcall wait
	ldi isopen,0
	rjmp loop


falseknock:
	dec temp2
	brne knockdetback
	rjmp noknock


;Timer 1 is ran to automatically close the lock if it's been open for about 15 minutes
t1int:
	push temp
	in temp,sreg
	push temp
	push temp2
	tst isopen ;Do we need this?
	breq nodoclose
	inc closectr ;Are we at 15 minutes already?
	cpi closectr,108
	brne t1end ;Nope...
	;Yes we are. Close the lock.
	ldi temp,80
autocloseloop:
	push temp
	rcall motorright
	pop temp
	dec temp
	brne autocloseloop
	ldi temp,0
	out portb,temp
	ldi isopen,0
nodoclose:
	ldi closectr,0
t1end:
	pop temp2
	pop temp
	out sreg,temp
	pop temp
	reti


;Gets a valid knock. Temp=response: 0 if no knock, 1 if a knock, 2 if invalid knock.
;Yes, there's magic inhere. 
getknock:
	push temp2
	
	ldi temp2,50
	
	;A knock should last a certain time. Check that.
knockdetback:
	ldi temp,255
knockdet1:
	sbis pind,knock
	 rjmp falseknock
	nop
	nop
	nop
	nop
	dec temp
	brne knockdet1

	;Wait a certain time to allow resonance in the door etc to die out
	ldi temp,250	
	rcall waitc
	ldi temp,250	
	rcall waitc
	ldi temp,250	
	rcall waitc

	;There shouldn't be any sound anymore; the knock should've ended. Check that.
	ldi temp2,30
knockdet2a:
	ldi temp,255
knockdet2:
	sbic pind,knock
	 rjmp invknock
	nop
	nop
	dec temp
	brne knockdet2
	dec temp2
	brne knockdet2a

	;If we arrive here, we're quite certain it was a knock.	
	pop temp2
	ldi temp,1
	ret
	
noknock:
	;No knock detected.
	ldi temp,0
	pop temp2
	ret

invknock:
	;We detected something, but it wasn't a knock.
	ldi temp,2
	pop temp2
	ret

;Routine to make the motor go one step to the left.
motorleft:
	ldi temp,0b0001
	rcall setmotor
	ldi temp,0b0101
	rcall setmotor
	ldi temp,0b0100
	rcall setmotor
	ldi temp,0b0110
	rcall setmotor
	ldi temp,0b0010
	rcall setmotor
	ldi temp,0b1010
	rcall setmotor
	ldi temp,0b1000
	rcall setmotor
	ldi temp,0b1001
	rcall setmotor
	ret

;Routine to make the motor go one step to the right.
motorright:
	ldi temp,0b1001
	rcall setmotor
	ldi temp,0b1000
	rcall setmotor
	ldi temp,0b1010
	rcall setmotor
	ldi temp,0b0010
	rcall setmotor
	ldi temp,0b0110
	rcall setmotor
	ldi temp,0b0100
	rcall setmotor
	ldi temp,0b0101
	rcall setmotor
	ldi temp,0b0001
	rcall setmotor
	ret

;Output motor values in temp and wait a while for the motor to settle.
setmotor:
	cbi portb,m1
	cbi portb,m2
	cbi portb,m3
	cbi portb,m4
	sbrc temp,0
	 sbi portb,m1
	sbrc temp,1
	 sbi portb,m2
	sbrc temp,2
	 sbi portb,m3
	sbrc temp,3
	 sbi portb,m4
	ldi temp,35
	rcall waitc
;	ldi temp,0
;	out portb,temp
;	ldi temp,15
;	rcall waitc

	ret

;Wait a certain amount of time indicated by temp
waitc:
	push temp
	push temp2
	rjmp w1

	
;Wait a fixed certain amount of time. Abort if a knock is detected.
wait2:
	push temp
	push temp2
	ldi temp,60
w21:
	ldi temp2,0
w22:
	sbic pind,knock
	 rjmp w2abort
	dec temp2
	brne w22

	dec temp
	brne w21
w2abort:
	pop temp2
	pop temp
	ret

;Wait a certain amount of time.
wait:
	push temp
	push temp2
	ldi temp,50
w1:
	ldi temp2,0
w2:
	nop
	dec temp2
	brne w2

	dec temp
	brne w1
	
	pop temp2
	pop temp
	
	ret